"""Central place for install-checker and guards for 'extras' dependencies.

The ``*_INSTALLED`` constansts reveal wether the dependency is installed with a supported version.
The :py:func:`install_guard` guard function is intended for use inside functions which need specific
extra packages installed.

Example usage:

.. code-block:: python

    from rstcheck_core import _extras

    if _extras.SPHINX_INSTALLED:
        import sphinx


    def print_sphinx_version():
        _extras.install_guard("sphinx")
        print(sphinx.version_info)
"""

from __future__ import annotations

import importlib
import importlib.metadata
import logging
import typing as t

logger = logging.getLogger(__name__)


ExtraDependencies = t.Literal["sphinx", "tomli"]
"""List of all dependencies installable through extras."""


class DependencyInfos(t.TypedDict):
    """Information about a dependency."""

    min_version: tuple[int, ...]
    extra: str


ExtraDependenciesInfos: dict[ExtraDependencies, DependencyInfos] = {
    "sphinx": DependencyInfos(min_version=(5, 0), extra="sphinx"),
    "tomli": DependencyInfos(min_version=(2, 0), extra="toml"),
}
"""Dependency map with their min. supported version and extra by which they can be installed."""


def is_installed_with_supported_version(package: ExtraDependencies) -> bool:
    """Check if the package is installed and has the minimum required version.

    :param package: Name of packge to check
    :return: Bool if package is installed with supported version
    """
    logger.debug(
        "Check if package is installed with supported version: '%s'.",
        package,
    )
    try:
        importlib.import_module(package)
    except ImportError:
        return False

    version: str = importlib.metadata.version(package)
    version_tuple = tuple(int(v) for v in version.split(".")[:3])

    return version_tuple >= ExtraDependenciesInfos[package]["min_version"]


SPHINX_INSTALLED = is_installed_with_supported_version("sphinx")
TOMLI_INSTALLED = is_installed_with_supported_version("tomli")


ExtraDependenciesInstalled: dict[ExtraDependencies, bool] = {
    "sphinx": SPHINX_INSTALLED,
    "tomli": TOMLI_INSTALLED,
}


def install_guard(package: ExtraDependencies) -> None:
    """Guard code that needs the ``package`` installed and throw :py:exc:`ModuleNotFoundError`.

    See example in module docstring.

    :param package: Name of packge to check
    :raises ModuleNotFoundError: When the package is not installed.
    """
    if ExtraDependenciesInstalled[package] is True:
        return

    extra = ExtraDependenciesInfos[package]

    msg = (
        f"No supported version of {package} installed. Install rstcheck with "
        f"{extra} extra (rstcheck[{extra}]) or install a supported version of {package} yourself."
    )
    raise ModuleNotFoundError(msg)


def install_guard_tomli(*, tomllib_imported: bool) -> None:
    """Specific version of :py:func:`install_guard` for ``tomli``.

    :param tomllib_imported: If tomllib is imported
    :raises ModuleNotFoundError: When ``tomli`` is not installed.
    """
    if tomllib_imported or ExtraDependenciesInstalled["tomli"] is True:
        return

    extra = ExtraDependenciesInfos["tomli"]

    msg = (
        f"tomllib could not be imported and no supported version of tomli installed. "
        f"Install rstcheck with {extra} extra (rstcheck[{extra}]) or install a "
        "supported version of tomli yourself."
    )
    raise ModuleNotFoundError(msg)



"""Runner of rstcheck_core."""

from __future__ import annotations

import logging
import multiprocessing
import os
import pathlib
import re
import sys
import typing as t

from . import _sphinx, checker, config, types

logger = logging.getLogger(__name__)


class RstcheckMainRunner:
    """Main runner of rstcheck_core."""

    def __init__(
        self,
        check_paths: list[pathlib.Path],
        rstcheck_config: config.RstcheckConfig,
        *,
        overwrite_config: bool = True,
    ) -> None:
        """Initialize the :py:class:`RstcheckMainRunner` with a base config.

        :param check_paths: Files to check.
        :param rstcheck_config: Base configuration config from e.g. the CLI.
        :param overwrite_config: If file config overwrites current config; defaults to True
        """
        self.config = rstcheck_config
        self.overwrite_config = overwrite_config
        if rstcheck_config.config_path:
            self.load_config_file(
                rstcheck_config.config_path,
                warn_unknown_settings=rstcheck_config.warn_unknown_settings or False,
            )

        self.check_paths = check_paths
        self._files_to_check: list[pathlib.Path] = []
        self._nonexisting_paths: list[pathlib.Path] = []
        self.update_file_list()

        pool_size = multiprocessing.cpu_count()
        # NOTE: Work around https://bugs.python.org/issue45077
        self._pool_size = pool_size if sys.platform != "win32" else min(pool_size, 61)

        self.errors: list[types.LintError] = []

    @property
    def files_to_check(self) -> list[pathlib.Path]:
        """List of files to check.

        This list is updated via the :py:meth:`RstcheckMainRunner.update_file_list` method.
        """
        return self._files_to_check

    @property
    def nonexisting_paths(self) -> list[pathlib.Path]:
        """List of paths which do not exist.

        This list is updated via the :py:meth:`RstcheckMainRunner.update_file_list` method.
        """
        return self._nonexisting_paths

    def load_config_file(
        self, config_path: pathlib.Path, *, warn_unknown_settings: bool = False
    ) -> None:
        """Load config from file and merge with current config.

        If the loaded file config overwrites the current config depends on the
        ``self.overwrite_config`` attribute set on initialization.

        :param config_path: Path to config file; can be directory or file
        :param warn_unknown_settings: If a warning should be logged for unknown settings in config
            file;
            defaults to :py:obj:`False`
        """
        logger.info("Load config file for main runner: '%s'.", config_path)
        file_config = config.load_config_file_from_path(
            config_path, warn_unknown_settings=warn_unknown_settings
        )

        if file_config is None:
            logger.warning("Config file was empty or not found.")
            return

        logger.debug(
            "Merging config from file into main config. File config is dominant: %s",
            self.overwrite_config,
        )
        self.config = config.merge_configs(
            self.config, file_config, config_add_is_dominant=self.overwrite_config
        )

    def update_file_list(self) -> None:
        """Update file path list with paths specified on initialization.

        Uses paths from ``RstcheckMainRunner.check_paths``, resolves all file paths and
        saves them in :py:attr:`RstcheckMainRunner.files_to_check`.

        If a given path does not exist, it is filtered out and saved in
        :py:attr:`RstcheckMainRunner.files_to_check`.

        Clear the current file list. Then get the file and directory paths specified with
        ``self.check_paths`` attribute set on initialization and search them for rst files
        to check. Add those files to the file list.
        """
        logger.debug("Updating list of files to check.")
        paths = list(self.check_paths)
        self._files_to_check = []

        if len(paths) == 1 and paths[0].name == "-":
            logger.info("'-' detected. Using stdin for input.'")
            self._files_to_check.append(paths[0])
            return

        paths = self._filter_nonexisting_paths(paths)

        def checkable_rst_file(f: pathlib.Path) -> bool:
            return f.is_file() and not f.name.startswith(".") and f.suffix.casefold() == ".rst"

        while paths:
            path = paths.pop(0)
            resolved_path = path.resolve()
            if self.config.recursive and resolved_path.is_dir():
                for root, directories, children in os.walk(path):
                    root_path = pathlib.Path(root)
                    paths += [
                        root_path / f
                        for f in children
                        if checkable_rst_file((root_path / f).resolve())
                    ]
                    directories[:] = [d for d in directories if not d.startswith(".")]
                continue

            if checkable_rst_file(resolved_path):
                self._files_to_check.append(path)

    def _filter_nonexisting_paths(self, paths: list[pathlib.Path]) -> list[pathlib.Path]:
        """Filter nonexisting paths out.

        If recursive is not active only files are allowed, else directories are also allowed.

        :param paths: List of paths to filter
        :return: Filtered path list
        """
        self._nonexisting_paths = []
        _paths = list(paths)

        for path in _paths:
            resolved_path = path.resolve()

            if resolved_path.is_file():
                continue

            if self.config.recursive and resolved_path.is_dir():
                continue

            _paths.remove(path)
            self._nonexisting_paths.append(path)

            if self.config.recursive:
                logger.warning(
                    "Path does not exist or is neither a file nor a directory: '%s'.",
                    path,
                )
                continue

            logger.warning(
                "Path does not exist or is not a file: '%s'.",
                path,
            )

        return _paths

    def _run_checks_sync(self) -> list[list[types.LintError]]:
        """Check all files from the file list syncronously and return the errors.

        :return: List of lists of errors found per file
        """
        logger.debug("Runnning checks synchronically.")
        with _sphinx.load_sphinx_if_available():
            return [
                checker.check_file(file, self.config, self.overwrite_config)
                for file in self._files_to_check
            ]

    def _run_checks_parallel(self) -> list[list[types.LintError]]:
        """Check all files from the file list in parallel and return the errors.

        :return: List of lists of errors found per file
        """
        logger.debug(
            "Runnning checks in parallel with pool size of %s.",
            self._pool_size,
        )
        with _sphinx.load_sphinx_if_available(), multiprocessing.Pool(self._pool_size) as pool:
            return pool.starmap(
                checker.check_file,
                [(file, self.config, self.overwrite_config) for file in self._files_to_check],
            )

    def _update_results(self, results: list[list[types.LintError]]) -> None:
        """Take results and update error cache.

        Result normally come from :py:meth:`RstcheckMainRunner._run_checks_sync` or
        :py:meth:`RstcheckMainRunner._run_checks_parallel`.
        :param results: List of lists of errors found
        """
        self.errors = []
        for errors in results:
            self.errors += errors

    def check(self) -> None:
        """Check all files in the file list and save the errors.

        Multiple files are run in parallel.

        A new call overwrite the old cached errors.
        """
        logger.info("Run checks for all files.")
        results = (
            self._run_checks_parallel()
            if len(self._files_to_check) > 1
            else self._run_checks_sync()
        )
        self._update_results(results)

    def print_result(self, output_file: t.TextIO | None = None) -> int:
        """Print all cached error messages and return exit code.

        :param output_file: file to print to; defaults to sys.stderr (if ``None``)
        :return: exit code 0 if no error is printed; 1 if any error is printed
        """
        if len(self.errors) == 0 and len(self._nonexisting_paths) == 0:
            print("Success! No issues detected.", file=output_file or sys.stdout)
            return 0

        err_msg_regex = re.compile(r"\([A-Z]+/[0-9]+\)")

        for error in self.errors:
            err_msg = error["message"]
            if not err_msg_regex.match(err_msg):
                err_msg = "(ERROR/3) " + err_msg

            message = f"{error['source_origin']}:{error['line_number']}: {err_msg}"

            print(message, file=output_file or sys.stderr)

        print("Error! Issues detected.", file=output_file or sys.stderr)
        return 1

    def run(self) -> int:  # pragma: no cover
        """Run checks, print error messages and return the result.

        :return: exit code 0 if no error is printed; 1 if any error is printed
        """
        logger.info("Run checks and print results.")
        self.check()
        return self.print_result()



"""Checking functionality."""

from __future__ import annotations

import contextlib
import copy
import doctest
import io
import json
import locale
import logging
import os
import pathlib
import re
import shlex
import subprocess
import sys
import tempfile
import typing as t
import warnings
import xml.etree.ElementTree

import docutils.core
import docutils.io
import docutils.nodes
import docutils.utils

from . import _docutils, _extras, _sphinx, config, inline_config, types

try:
    import yaml

    yaml_imported = True
except ImportError:
    yaml_imported = False


logger = logging.getLogger(__name__)


EXCEPTION_LINE_NO_REGEX = re.compile(r": line\s+([0-9]+)[^:]*$")
DOCTEST_LINE_NO_REGEX = re.compile(r"line ([0-9]+)")
MARKDOWN_LINK_REGEX = re.compile(r"\[[^\]]+\]\([^\)]+\)")


def check_file(
    source_file: pathlib.Path,
    rstcheck_config: config.RstcheckConfig,
    overwrite_with_file_config: bool = True,  # noqa: FBT001,FBT002
) -> list[types.LintError]:
    """Check the given file for issues.

    On every call docutils' caches for roles and directives are cleared by reloading their modules.

    :param source_file: Path to file to check
    :param rstcheck_config: Main configuration of the application
    :param overwrite_with_file_config: If the loaded file config should overwrite the
        ``rstcheck_config``;
        defaults to :py:obj:`True`
    :return: A list of found issues
    """
    logger.info("Check file'%s'", source_file)
    run_config = _load_run_config(
        source_file.parent, rstcheck_config, overwrite_config=overwrite_with_file_config
    )
    ignore_dict = _create_ignore_dict_from_config(run_config)

    source = _get_source(source_file)

    _docutils.clean_docutils_directives_and_roles_cache()

    with _sphinx.load_sphinx_if_available():
        return list(
            check_source(
                source,
                source_file=source_file,
                ignores=ignore_dict,
                report_level=run_config.report_level or config.DEFAULT_REPORT_LEVEL,
                warn_unknown_settings=run_config.warn_unknown_settings or False,
            )
        )


def _load_run_config(
    source_file_dir: pathlib.Path,
    rstcheck_config: config.RstcheckConfig,
    *,
    overwrite_config: bool = True,
) -> config.RstcheckConfig:
    """Load file specific config file and create run config.

    If the ``rstcheck_config`` does not contain a ``config_path`` the ``source_file_dir`` directory
    tree is searched for a config file to load and merge into the ``rstcheck_config``. The merge
    strategy is set via ``overwrite_config``.

    :param source_file_dir: Directory of the current file to check
    :param rstcheck_config: Main configuration of the application
    :param overwrite_config: If the loaded config should overwrite the ``rstcheck_config``;
        defaults to :py:obj:`True`
    :return: Merged config
    """
    if rstcheck_config.config_path is not None:
        return rstcheck_config

    file_config = config.load_config_file_from_dir_tree(source_file_dir)

    if file_config is None:
        return rstcheck_config

    return config.merge_configs(
        copy.copy(rstcheck_config), file_config, config_add_is_dominant=overwrite_config
    )


def _get_source(source_file: pathlib.Path) -> str:
    """Get source from file or stdin.

    If the file name is "-" then stdin is read for input instead of a file.

    :param source_file: File path to read contents from
    :return: Loaded content
    """
    if source_file.name == "-":
        logger.info("Load source from stdin.")
        return sys.stdin.read()

    resolved_file_path = source_file.resolve()
    with contextlib.closing(docutils.io.FileInput(source_path=resolved_file_path)) as input_file:
        return input_file.read()


def _replace_ignored_substitutions(source: str, ignore_substitutions: list[str]) -> str:
    """Replace rst substitutions from the ignore list with a dummy.

    :param source: Source to replace substitutions in
    :param ignore_substitutions: Substitutions to replace with dummy
    :return: Cleaned source
    """
    for substitution in ignore_substitutions:
        source = source.replace(f"|{substitution}|", f"x{substitution}x")
    return source


def _create_ignore_dict_from_config(rstcheck_config: config.RstcheckConfig) -> types.IgnoreDict:
    """Extract ignore settings from config and create a :py:class:`rstcheck_core.types.IgnoreDict`.

    :param rstcheck_config: Config to extract ignore settings from
    :return: :py:class:`rstcheck_core.types.IgnoreDict`
    """
    return types.construct_ignore_dict(
        messages=rstcheck_config.ignore_messages,
        languages=rstcheck_config.ignore_languages,
        directives=rstcheck_config.ignore_directives,
        roles=rstcheck_config.ignore_roles,
        substitutions=rstcheck_config.ignore_substitutions,
    )


def check_source(
    source: str,
    source_file: types.SourceFileOrString | None = None,
    ignores: types.IgnoreDict | None = None,
    report_level: config.ReportLevel = config.DEFAULT_REPORT_LEVEL,
    *,
    warn_unknown_settings: bool = False,
) -> types.YieldedLintError:
    """Check the given rst source for issues.

    :param source_file: Path to file the source comes from if it comes from a file;
        defaults to :py:obj:`None`
    :param ignores: Ignore information; defaults to :py:obj:`None`
    :param report_level: Report level; defaults to
        :py:data:`rstcheck_core.config.DEFAULT_REPORT_LEVEL`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :return: :py:obj:`None`
    :yield: Found issues
    """
    source_origin: types.SourceFileOrString = source_file or "<string>"
    if isinstance(source_origin, pathlib.Path) and source_origin.name == "-":
        source_origin = "<stdin>"
    logger.info("Check source from '%s'", source_origin)
    ignores = ignores or types.construct_ignore_dict()
    ignores["directives"].extend(
        inline_config.find_ignored_directives(
            source, source_origin, warn_unknown_settings=warn_unknown_settings
        )
    )
    ignores["roles"].extend(
        inline_config.find_ignored_roles(
            source, source_origin, warn_unknown_settings=warn_unknown_settings
        )
    )
    ignores["substitutions"].extend(
        inline_config.find_ignored_substitutions(
            source, source_origin, warn_unknown_settings=warn_unknown_settings
        )
    )
    ignores["languages"].extend(
        inline_config.find_ignored_languages(
            source, source_origin, warn_unknown_settings=warn_unknown_settings
        )
    )

    source = _replace_ignored_substitutions(source, ignores["substitutions"])

    _docutils.register_code_directive(
        ignore_code_directive="code" in ignores["directives"],
        ignore_codeblock_directive="code-block" in ignores["directives"],
        ignore_sourcecode_directive="sourcecode" in ignores["directives"],
    )

    _docutils.ignore_directives_and_roles(ignores["directives"] or [], ignores["roles"] or [])

    if _extras.SPHINX_INSTALLED:
        _sphinx.load_sphinx_ignores()

    writer = _CheckWriter(source, source_origin, ignores, report_level)

    string_io = io.StringIO()

    # This is a hack to avoid false positive from docutils (#23). docutils mistakes BOMs for actual
    # visible letters. This results in the "underline too short" warning firing.
    # This is tested in the CLI integration tests with the `testing/examples/good/bom.rst` file.
    with contextlib.suppress(UnicodeError):
        source = source.encode("utf-8").decode("utf-8-sig")

    with contextlib.suppress(docutils.utils.SystemMessage):
        # Sphinx will sometimes throw an `AttributeError` trying to access
        # "self.state.document.settings.env". Ignore this for now until we
        # figure out a better approach.
        # https://github.com/rstcheck/rstcheck-core/issues/3
        try:
            docutils.core.publish_string(
                source,
                writer=writer,
                source_path=str(source_origin),
                settings_overrides={
                    "halt_level": 5,
                    "report_level": report_level.value,
                    "warning_stream": string_io,
                },
            )
        except AttributeError:
            if not _extras.SPHINX_INSTALLED:
                raise
            logger.warning(
                "An `AttributeError` error occured. This is most probably due to a code block "
                "directive (code/code-block/sourcecode) without a specified language. "
                "This may result in a false negative for source: '%s'. "
                "The reason can also be another directive. "
                "For more information see the FAQ (https://rstcheck-core.rtfd.io/en/latest/faq) "
                "or the corresponding github issue: "
                "https://github.com/rstcheck/rstcheck-core/issues/3.",
                source_origin,
            )

    yield from _run_code_checker_and_filter_errors(writer.checkers, ignores["messages"])

    rst_errors = string_io.getvalue().strip()

    if not rst_errors:
        return

    yield from _parse_and_filter_rst_errors(rst_errors, source_origin, ignores["messages"])


def _run_code_checker_and_filter_errors(
    checker_list: list[types.CheckerRunFunction],
    ignore_messages: t.Pattern[str] | None = None,
) -> types.YieldedLintError:
    """Run all code block checker functions.

    :param checker_list: List of code block checker functions
    :param ignore_messages: Regex for ignoring error messages;
        defaults to :py:obj:`None`
    :return: :py:obj:`None`
    :yield: Filtered :py:class:`rstcheck_core.types.LintError` s from run checker function
    """
    for checker in checker_list:
        for lint_error in checker():
            if ignore_messages and ignore_messages.search(lint_error["message"]):
                continue
            yield lint_error


def _parse_and_filter_rst_errors(
    rst_errors: str,
    source_origin: types.SourceFileOrString,
    ignore_messages: t.Pattern[str] | None = None,
) -> types.YieldedLintError:
    """Parse rst errors and yield filtered :py:class:`rstcheck_core.types.LintError`.

    :param rst_errors: String with rst errors
    :param source_origin: Origin of the source with the errors
    :param ignore_messages: Regex for ignoring error messages;
        defaults to :py:obj:`None`
    :return: :py:obj:`None`
    :yield: Parsed and filtered :py:class:`rstcheck_core.types.LintError` s
    """
    for message in rst_errors.splitlines():
        with contextlib.suppress(ValueError):
            if ignore_messages and ignore_messages.search(message):
                continue
            yield _parse_gcc_style_error_message(
                message, source_origin=source_origin, has_column=False
            )


class _CheckWriter(docutils.writers.Writer):
    """Runs CheckTranslator on code blocks."""

    def __init__(
        self,
        source: str,
        source_origin: types.SourceFileOrString,
        ignores: types.IgnoreDict | None = None,
        report_level: config.ReportLevel = config.DEFAULT_REPORT_LEVEL,
        *,
        warn_unknown_settings: bool = False,
    ) -> None:
        """Inititalize :py:class:`_CheckWriter`.

        :param source: Rst source to check
        :param source_origin: Path to file the source comes from
        :param ignores: Ignore information; defaults to :py:obj:`None`
        :param report_level: Report level;
            defaults to :py:data:`rstcheck_core.config.DEFAULT_REPORT_LEVEL`
        :param warn_unknown_settings: If a warning should be logged for unknown settings in config
            file;
            defaults to :py:obj:`False`
        """
        docutils.writers.Writer.__init__(self)
        self.checkers: list[types.CheckerRunFunction] = []
        self.source = source
        self.source_origin = source_origin
        self.ignores = ignores
        self.report_level = report_level
        self.warn_unknown_settings = warn_unknown_settings

    def translate(self) -> None:
        """Run CheckTranslator."""
        visitor = _CheckTranslator(
            self.document,
            source=self.source,
            source_origin=self.source_origin,
            ignores=self.ignores,
            report_level=self.report_level,
            warn_unknown_settings=self.warn_unknown_settings,
        )
        self.document.walkabout(visitor)
        self.checkers += visitor.checkers


class _CheckTranslator(docutils.nodes.NodeVisitor):
    """Visits code blocks and checks for syntax errors in code."""

    def __init__(  # noqa: PLR0913
        self,
        document: docutils.nodes.document,
        source: str,
        source_origin: types.SourceFileOrString,
        ignores: types.IgnoreDict | None = None,
        report_level: config.ReportLevel = config.DEFAULT_REPORT_LEVEL,
        *,
        warn_unknown_settings: bool = False,
    ) -> None:
        """Inititalize :py:class:`_CheckTranslator`.

        :param document: Document node
        :param source: Rst source to check
        :param source_origin: Path to file the source comes from
        :param ignores: Ignore information; defaults to :py:obj:`None`
        :param report_level: Report level;
            defaults to :py:data:`rstcheck_core.config.DEFAULT_REPORT_LEVEL`
        :param warn_unknown_settings: If a warning should be logged for unknown settings in config
            file;
            defaults to :py:obj:`False`
        """
        docutils.nodes.NodeVisitor.__init__(self, document)
        self.checkers: list[types.CheckerRunFunction] = []
        self.source = source
        self.source_origin = source_origin
        self.ignores = ignores or types.construct_ignore_dict()
        self.report_level = report_level
        self.warn_unknown_settings = warn_unknown_settings
        self.code_block_checker = CodeBlockChecker(
            source_origin, ignores, report_level, warn_unknown_settings=warn_unknown_settings
        )
        self.code_block_ignore_lines = list(
            inline_config.find_code_block_ignore_lines(
                source=self.source,
                source_origin=self.source_origin,
                warn_unknown_settings=self.warn_unknown_settings,
            )
        )

    def visit_doctest_block(self, node: docutils.nodes.Element) -> None:
        """Add check for syntax of doctest.

        :param node: The doctest node
        """
        if "doctest" in self.ignores["languages"]:
            return

        self._add_check(
            node=node,
            run=self.code_block_checker.create_checker(node.rawsource, "doctest"),
            language="doctest",
            is_code_node=False,
        )

    def visit_literal_block(self, node: docutils.nodes.Element) -> None:
        """Add check for syntax of code block.

        :param node: The code block node
        :raises docutils.nodes.SkipNode: After a check was added or nothing to do
        """
        # For "..code-block:: language"
        language = node.get("language")
        is_code_node = False
        if not language:
            # For "..code:: language"
            is_code_node = True
            classes = node.get("classes")
            if "code" not in classes:
                return
            language = classes[-1]

        directive_line = _get_code_block_directive_line(node, self.source)
        if directive_line is None:
            logger.warning(
                "Could not find line for literal block directive. "
                "This could be due to an indented code block. "
                "This message only warns about the missing line number "
                "and is no error of itself. "
                "Source: '%s'%s",
                self.source_origin,
                f" at line {node.line}" if node.line is not None else "",
            )
            return

        if directive_line - 1 in self.code_block_ignore_lines:
            logger.debug(
                "Skipping code-block due to skip comment. Source: '%s' at line %s",
                self.source_origin,
                node.line,
            )
            return

        if language in self.ignores["languages"]:
            return

        if language == "doctest" or (
            language == "python" and node.rawsource.lstrip().startswith(">>> ")
        ):
            self.visit_doctest_block(node)
            raise docutils.nodes.SkipNode

        if self.code_block_checker.language_is_supported(language):
            run = self.code_block_checker.create_checker(node.rawsource, language)
            self._add_check(node=node, run=run, language=language, is_code_node=is_code_node)

        raise docutils.nodes.SkipNode

    def visit_paragraph(self, node: docutils.nodes.Element) -> None:
        """Check syntax of reStructuredText.

        :param node: The rst node
        """
        find = MARKDOWN_LINK_REGEX.search(node.rawsource)
        if find is not None:
            self.document.reporter.warning(
                "(rst) Link is formatted in Markdown style.", base_node=node
            )

    def _add_check(
        self,
        node: docutils.nodes.Element,
        run: types.CheckerRunFunction,
        language: str,
        *,
        is_code_node: bool,
    ) -> None:
        """Add node checker that will be run.

        :param node: The node to check
        :param run: The runner function that checks the node
        :param language: The language of the node
        :param is_code_node: If it is a code block node
        """

        def run_check() -> types.YieldedLintError:
            """Yield found issues."""
            all_results = run()
            if all_results is not None:
                if all_results:
                    for result in all_results:
                        error_offset = result["line_number"] - 1

                        line_number = getattr(node, "line", None)
                        if line_number is not None:
                            yield types.LintError(
                                source_origin=result["source_origin"],
                                line_number=_beginning_of_code_block(
                                    node=node,
                                    line_number=line_number,
                                    full_contents=self.source,
                                    is_code_node=is_code_node,
                                )
                                + error_offset,
                                message=f"({language}) {result['message']}",
                            )
                else:
                    yield types.LintError(
                        source_origin=self.source_origin, line_number=0, message="unknown error"
                    )

        self.checkers.append(run_check)

    def unknown_visit(self, node: docutils.nodes.Node) -> None:
        """Ignore."""

    def unknown_departure(self, node: docutils.nodes.Node) -> None:
        """Ignore."""


def _beginning_of_code_block(
    node: docutils.nodes.Element, line_number: int, full_contents: str, *, is_code_node: bool
) -> int:
    """Get line number of beginning of code block.

    :param node: The code block node
    :param line_number: The current line number
    :param full_contents: The node's contents
    :param is_code_node: If it is a code block node
    :return: First fine number of the block
    """
    if _extras.SPHINX_INSTALLED and not is_code_node:
        sphinx_code_block_delta = -1
        delta = len(node.non_default_attributes())
        current_line_contents = full_contents.splitlines()[line_number:]
        blank_lines = next((i for (i, x) in enumerate(current_line_contents) if x), 0)
        return line_number + delta - 1 + blank_lines - 1 + sphinx_code_block_delta

    lines = full_contents.splitlines()
    code_block_length = len(node.rawsource.splitlines())

    with contextlib.suppress(IndexError):
        # Case where there are no extra spaces.
        if lines[line_number - 1].strip():
            return line_number - code_block_length + 1

    # The offsets are wrong if the RST text has multiple blank lines after
    # the code block. This is a workaround.
    for line_no in range(line_number, 1, -1):
        if lines[line_no - 2].strip():
            break

    return line_no - code_block_length


CODE_BLOCK_RE = re.compile(r"\.\. code::|\.\. code-block::|\.\. sourcecode::")


def _get_code_block_directive_line(node: docutils.nodes.Element, full_contents: str) -> int | None:
    """Find line of code block directive.

    :param node: The code block node
    :param full_contents: The node's contents
    :return: Line of code block directive or :py:obj:`None`
    """
    line_number = node.line
    if line_number is None:
        return None

    if _extras.SPHINX_INSTALLED:
        return line_number

    lines = full_contents.splitlines()
    for line_no in range(line_number, 1, -1):
        if CODE_BLOCK_RE.match(lines[line_no - 2].strip()) is not None:
            return line_no - 1

    return None


class CodeBlockChecker:
    """Checker for code blockes with different languages."""

    def __init__(
        self,
        source_origin: types.SourceFileOrString,
        ignores: types.IgnoreDict | None = None,
        report_level: config.ReportLevel = config.DEFAULT_REPORT_LEVEL,
        *,
        warn_unknown_settings: bool = False,
    ) -> None:
        """Inititalize CodeBlockChecker.

        :param source_origin: Path to file the source comes from
        :param ignores: Ignore information; defaults to :py:obj:`None`
        :param report_level: Report level;
            defaults to :py:data:`rstcheck_core.config.DEFAULT_REPORT_LEVEL`
        :param warn_unknown_settings: If a warning should be logged for unknown settings in config
            file;
            defaults to :py:obj:`False`
        """
        self.source_origin = source_origin
        self.ignores = ignores
        self.report_level = report_level
        self.warn_unknown_settings = warn_unknown_settings

    def language_is_supported(self, language: str) -> bool:
        """Check if given language can be checked.

        :param language: Language to check
        :return: If langauge can be checked
        """
        return getattr(self, f"check_{language}", None) is not None

    def create_checker(self, source_code: str, language: str) -> types.CheckerRunFunction:
        """Create a checker function for the given source and language.

        :param source: Source code to check
        :param language: Language of the source code
        :return: Checker function
        """
        return lambda: self.check(source_code, language)

    def check(self, source_code: str, language: str) -> types.YieldedLintError:
        """Call the appropiate checker function for the given langauge to check given source.

        :param source: Source code to check
        :param language: Language of the source code
        :return: :py:obj:`None` if language is not supported
        :yield: Found issues
        """
        checker_function = t.Callable[[str], types.YieldedLintError]
        checker: checker_function | None = getattr(self, f"check_{language}", None)
        if checker is None:
            return None

        yield from checker(source_code)
        return None

    def check_python(self, source_code: str) -> types.YieldedLintError:
        """Check python source for syntax errors.

        :param source: Python source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check python source.")
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("error", SyntaxWarning)
                compile(source_code, "<string>", "exec")
        except SyntaxError as exception:
            yield types.LintError(
                source_origin=self.source_origin,
                line_number=int(exception.lineno or 0),
                message=exception.msg,
            )

    def check_json(self, source_code: str) -> types.YieldedLintError:
        """Check JSON source for syntax errors.

        :param source: JSON source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check JSON source.")
        try:
            json.loads(source_code)
        except ValueError as exception:
            message = f"{exception}"
            found = EXCEPTION_LINE_NO_REGEX.search(message)
            line_number = int(found.group(1)) if found else 0

            yield types.LintError(
                source_origin=self.source_origin, line_number=int(line_number), message=message
            )

    def check_yaml(self, source_code: str) -> types.YieldedLintError:
        """Check YAML source for syntax errors.

        :param source: JSON source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        if not yaml_imported:
            logger.debug("PyYAML is not installed, ignoring YAML source.")
            return
        logger.debug("Check YAML source.")
        try:
            yaml.safe_load(source_code)
        except yaml.error.YAMLError as exception:
            message = f"{exception}"
            found = EXCEPTION_LINE_NO_REGEX.search(message)
            line_number = int(found.group(1)) if found else 0

            yield types.LintError(
                source_origin=self.source_origin, line_number=int(line_number), message=message
            )

    def check_xml(self, source_code: str) -> types.YieldedLintError:
        """Check XML source for syntax errors.

        :param source: XML source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check XML source.")
        try:
            xml.etree.ElementTree.fromstring(source_code)  # noqa: S314
        except xml.etree.ElementTree.ParseError as exception:
            message = f"{exception}"
            found = EXCEPTION_LINE_NO_REGEX.search(message)
            line_number = int(found.group(1)) if found else 0

            yield types.LintError(
                source_origin=self.source_origin, line_number=int(line_number), message=message
            )

    def check_rst(self, source_code: str) -> types.YieldedLintError:
        """Check nested rst source for syntax errors.

        :param source: rst source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check RST source.")
        yield from check_source(
            source_code,
            source_file=self.source_origin,
            ignores=self.ignores,
            report_level=self.report_level,
            warn_unknown_settings=self.warn_unknown_settings,
        )

    def check_doctest(self, source_code: str) -> types.YieldedLintError:
        """Check doctest source for syntax errors.

        This does not run the test as that would be unsafe. Nor does this
        check the Python syntax in the doctest. That could be purposely
        incorrect for testing purposes.

        :param source: XML source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check doctest source.")
        parser = doctest.DocTestParser()
        try:
            parser.parse(source_code)
        except ValueError as exception:
            message = f"{exception}"
            match = DOCTEST_LINE_NO_REGEX.match(message)
            if match:
                yield types.LintError(
                    source_origin=self.source_origin,
                    line_number=int(match.group(1)),
                    message=message,
                )

    def check_bash(self, source_code: str) -> types.YieldedLintError:
        """Check bash source for syntax errors.

        :param source: bash source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check bash source.")
        result = self._run_in_subprocess(source_code, ".bash", ["bash", "-n"])

        if result:
            (output, filename) = result
            prefix = str(filename) + ": line "
            for line in output.splitlines():
                if not line.startswith(prefix):  # pragma: no cover # NOTE: Case not reproducible
                    continue
                message = line[len(prefix) :]
                split_message = message.split(":", 1)
                yield types.LintError(
                    source_origin=self.source_origin,
                    line_number=int(split_message[0]) - 1,
                    message=split_message[1].strip(),
                )

    def check_c(self, source_code: str) -> types.YieldedLintError:
        """Check C source for syntax errors.

        :param source: C source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check C source.")
        return self._gcc_checker(
            source_code,
            ".c",
            [
                os.getenv("CC", "gcc"),
                *shlex.split(os.getenv("CFLAGS", "")),
                *shlex.split(os.getenv("CPPFLAGS", "")),
                "-I.",
                "-I..",
            ],
        )

    def check_cpp(self, source_code: str) -> types.YieldedLintError:
        """Check C++ source for syntax errors.

        :param source: C++ source code to check
        :return: :py:obj:`None`
        :yield: Found issues
        """
        logger.debug("Check C++ source.")
        yield from self._gcc_checker(
            # Add a newline to ignore "no newline at end of file" errors
            # that are reported using clang (e.g. on macOS).
            source_code + "\n",
            ".cpp",
            [
                os.getenv("CXX", "g++"),
                *shlex.split(os.getenv("CXXFLAGS", "")),
                *shlex.split(os.getenv("CPPFLAGS", "")),
                "-I.",
                "-I..",
            ],
        )

    def _gcc_checker(
        self, source_code: str, filename_suffix: str, arguments: list[str]
    ) -> types.YieldedLintError:
        """Check code blockes using gcc (Helper function).

        :param source_code: Source code to check
        :param filename_suffix: File suffix for language of the source code
        :param arguments: Command and arguments to run
        :return: :py:obj:`None`
        :yield: Found issues
        """
        result = self._run_in_subprocess(
            source_code, filename_suffix, [*arguments, "-pedantic", "-fsyntax-only"]
        )

        if result:
            (output, temp_file_name) = result
            for line in output.splitlines():
                try:
                    yield _parse_gcc_style_error_message(
                        line, source_origin=self.source_origin, temp_file_name=temp_file_name
                    )
                except ValueError:
                    continue

    def _run_in_subprocess(
        self,
        code: str,
        filename_suffix: str,
        arguments: list[str],
    ) -> tuple[str, pathlib.Path] | None:
        """Run checker in a subprocess (Helper function).

        :param source_code: Source code to check
        :param filename_suffix: File suffix for language of the source code
        :param arguments: Command and arguments to run
        :return: :py:obj:`None` if no issues were found else a tuple of the stderr and temp-file
            name
        """
        encoding = locale.getpreferredencoding() or sys.getdefaultencoding()

        source_origin_path = self.source_origin
        if isinstance(source_origin_path, str):
            source_origin_path = pathlib.Path(source_origin_path)

        # NOTE: On windows a file cannot be opened twice.
        # Therefore close it before using it in subprocess.
        temporary_file = tempfile.NamedTemporaryFile(
            mode="wb", suffix=filename_suffix, delete=False
        )
        temporary_file_path = pathlib.Path(temporary_file.name)
        try:
            temporary_file.write(code.encode("utf-8"))
            temporary_file.flush()
            temporary_file.close()

            subprocess.run(
                [*arguments, temporary_file.name],  # noqa: S603
                capture_output=True,
                cwd=source_origin_path.parent,
                check=True,
            )
        except subprocess.CalledProcessError as exc:
            return (exc.stderr.decode(encoding), temporary_file_path)
        else:
            return None
        finally:
            temporary_file_path.unlink()


def _parse_gcc_style_error_message(
    message: str,
    source_origin: types.SourceFileOrString,
    *,
    temp_file_name: pathlib.Path | None = None,
    has_column: bool = True,
) -> types.LintError:
    """Parse GCC-style error message.

    Return (line_number, message). Raise ValueError if message cannot be
    parsed.

    :param message: Message to parse
    :param filename: File the code block producing the errors comes from
    :param temp_file_name: File the message is associated with
    :param has_column: The the message has a column number; defaults to :py:obj:`True`
    :raises ValueError: If the message cannot be parsed
    :return: Parsed message
    """
    colons = 2 if has_column else 1
    prefix = str(temp_file_name or source_origin) + ":"
    if not message.startswith(prefix):
        logger.debug("Skipping unparsable message: '%s'.", message)
        msg = "Message cannot be parsed."
        raise ValueError(msg)
    message = message[len(prefix) :]
    split_message = message.split(":", colons)
    line_number = int(split_message[0])
    return types.LintError(
        source_origin=source_origin, line_number=line_number, message=split_message[colons].strip()
    )



"""Rstcheck configuration functionality."""

from __future__ import annotations

import configparser
import contextlib
import enum
import logging
import pathlib
import re
import typing as t

import pydantic

from . import _extras

tomllib_imported = False
try:
    import tomllib

    tomllib_imported = True
except ModuleNotFoundError:
    if _extras.TOMLI_INSTALLED:  # pragma: no cover
        import tomli as tomllib  # type: ignore[import-not-found,no-redef]


logger = logging.getLogger(__name__)


CONFIG_FILES = [".rstcheck.cfg", "setup.cfg"]
"""Supported default config files."""
if _extras.TOMLI_INSTALLED or tomllib_imported:  # pragma: no cover
    CONFIG_FILES = [".rstcheck.cfg", "pyproject.toml", "setup.cfg"]


class ReportLevel(enum.Enum):
    """Report levels supported by docutils."""

    INFO = 1
    WARNING = 2
    ERROR = 3
    SEVERE = 4
    NONE = 5


ReportLevelMap = {
    "info": 1,
    "warning": 2,
    "error": 3,
    "severe": 4,
    "none": 5,
}
"""Map docutils report levels in text form to numbers."""


DEFAULT_REPORT_LEVEL = ReportLevel.INFO
"""Default report level."""


def _split_str_validator(value: t.Any) -> list[str] | None:  # noqa: ANN401
    """Validate and parse strings and string-lists.

    Comma separated strings are split into a list.

    :param value: Value to validate
    :raises ValueError: If not a :py:class:`str` or :py:class:`list` of :py:class:`str`
    :return: List of strings
    """
    if value is None:
        return None

    if isinstance(value, str):
        return [v.strip() for v in value.split(",") if v.strip()]

    if isinstance(value, list) and all(isinstance(v, str) for v in value):
        return [v.strip() for v in value if v.strip()]

    msg = "Not a string or list of strings"
    raise TypeError(msg)


class RstcheckConfigFile(pydantic.BaseModel):
    """Rstcheck config file.

    :raises ValueError: If setting has incorrect value or type
    :raises pydantic.ValidationError: If setting is not parsable into correct type
    """

    report_level: t.Optional[ReportLevel] = None  # noqa: UP007
    ignore_directives: t.Optional[t.List[str]] = None  # noqa: UP007,UP006
    ignore_roles: t.Optional[t.List[str]] = None  # noqa: UP007,UP006
    ignore_substitutions: t.Optional[t.List[str]] = None  # noqa: UP007,UP006
    ignore_languages: t.Optional[t.List[str]] = None  # noqa: UP007,UP006
    ignore_messages: t.Optional[t.Pattern[str]] = None  # noqa: UP007

    @pydantic.field_validator("report_level", mode="before")
    @classmethod
    def valid_report_level(cls, value: t.Any) -> ReportLevel | None:  # noqa: ANN401
        """Validate the report_level setting.

        :param value: Value to validate
        :raises ValueError: If ``value`` is not a valid docutils report level
        :return: Instance of :py:class:`ReportLevel` or None if emptry string.
        """
        if value is None:
            return None

        if isinstance(value, ReportLevel):
            return value

        if value == "":
            return DEFAULT_REPORT_LEVEL

        if isinstance(value, bool):
            msg = "Invalid report level"
            raise TypeError(msg)

        if isinstance(value, str):
            if value.casefold() in set(ReportLevelMap):
                return ReportLevel(ReportLevelMap[value.casefold()])

            with contextlib.suppress(ValueError):
                value = int(value)

        max_report_lvl = 5
        min_report_lvl = 1
        if isinstance(value, int) and min_report_lvl <= value <= max_report_lvl:
            return ReportLevel(value)

        msg = "Invalid report level"
        raise TypeError(msg)

    @pydantic.field_validator(
        "ignore_directives",
        "ignore_roles",
        "ignore_substitutions",
        "ignore_languages",
        mode="before",
    )
    @classmethod
    def split_str(cls, value: t.Any) -> list[str] | None:  # noqa: ANN401
        """Validate and parse the following ignore_* settings.

        - ignore_directives
        - ignore_roles
        - ignore_substitutions
        - ignore_languages

        Comma separated strings are split into a list.

        :param value: Value to validate
        :raises ValueError: If not a :py:class:`str` or :py:class:`list` of :py:class:`str`
        :return: List of things to ignore in the respective category
        """
        return _split_str_validator(value)

    @pydantic.field_validator("ignore_messages", mode="before")
    @classmethod
    def join_regex_str(cls, value: t.Any) -> str | t.Pattern[str] | None:  # noqa: ANN401
        """Validate and concatenate the ignore_messages setting to a RegEx string.

        If a list ist given, the entries are concatenated with "|" to create an or RegEx.

        :param value: Value to validate
        :raises ValueError: If not a :py:class:`str` or :py:class:`list` of :py:class:`str`
        :return: A RegEx string with messages to ignore or :py:class:`typing.Pattern` if it is one
            already
        """
        if value is None:
            return None

        if isinstance(value, re.Pattern):
            return value

        if isinstance(value, list) and all(isinstance(v, str) for v in value):
            return r"|".join(value)

        if isinstance(value, str):
            return value

        msg = "Not a string or list of strings"
        raise TypeError(msg)


class RstcheckConfig(RstcheckConfigFile):
    """Rstcheck config.

    :raises ValueError: If setting has incorrect value or type
    :raises pydantic.ValidationError: If setting is not parsable into correct type
    """

    config_path: t.Optional[pathlib.Path] = None  # noqa: UP007
    recursive: t.Optional[bool] = None  # noqa: UP007
    warn_unknown_settings: t.Optional[bool] = None  # noqa: UP007


class _RstcheckConfigINIFile(pydantic.BaseModel):
    """Type for [rstcheck] section in INI file.

    The types apply to the file's data before the parsing by :py:class:`RstcheckConfig` is done.

    :raises pydantic.ValidationError: If setting is not parsable into correct type
    """

    report_level: t.Union[str, int, None] = None  # noqa: UP007
    ignore_directives: t.Optional[str] = None  # noqa: UP007
    ignore_roles: t.Optional[str] = None  # noqa: UP007
    ignore_substitutions: t.Optional[str] = None  # noqa: UP007
    ignore_languages: t.Optional[str] = None  # noqa: UP007
    ignore_messages: t.Optional[str] = None  # noqa: UP007


def _load_config_from_ini_file(
    ini_file: pathlib.Path,
    *,
    log_missing_section_as_warning: bool = True,
    warn_unknown_settings: bool = False,
) -> RstcheckConfigFile | None:
    """Load, parse and validate rstcheck config from a ini file.

    :param ini_file: INI file to load config from
    :param log_missing_section_as_warning: If a missing [tool.rstcheck] section should be logged at
        WARNING (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level;
        defaults to :py:obj:`True`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :raises FileNotFoundError: If the file is not found
    :return: instance of :py:class:`RstcheckConfigFile` or :py:class:`None` on missing config
        section
        or ``NONE`` is passed as the config path.
    """
    logger.debug("Try loading config from INI file: '%s'", ini_file)

    if ini_file.name == "NONE":
        logger.info("Config path is set to 'NONE'. No config file is loaded.")
        return None

    resolved_file = ini_file.resolve()

    if not resolved_file.is_file():
        msg = f"{resolved_file}"
        raise FileNotFoundError(msg)

    parser = configparser.ConfigParser()
    parser.read(resolved_file)

    if not parser.has_section("rstcheck"):
        if log_missing_section_as_warning:
            logger.warning(
                "Config file has no [rstcheck] section: '%s'.",
                ini_file,
            )
            return None
        logger.info(
            "Config file has no [rstcheck] section: '%s'.",
            ini_file,
        )
        return None

    config_values_raw = dict(parser.items("rstcheck"))
    if warn_unknown_settings:
        known_settings = _RstcheckConfigINIFile().model_dump().keys()
        unknown = [s for s in config_values_raw if s not in known_settings]
        if unknown:
            logger.warning(
                "Unknown setting(s) %s found in file: '%s'.",
                unknown,
                ini_file,
            )

    config_values_checked = _RstcheckConfigINIFile(**config_values_raw)
    return RstcheckConfigFile(**config_values_checked.model_dump())


class _RstcheckConfigTOMLFile(
    pydantic.BaseModel,
):
    """Type for [tool.rstcheck] section in TOML file.

    The types apply to the file's data before the parsing by :py:class:`RstcheckConfig` is done.

    :raises pydantic.ValidationError: If setting is not parsable into correct type
    """

    report_level: t.Union[str, int, None] = None  # noqa: UP007
    ignore_directives: t.Optional[t.List[str]] = None  # noqa: UP006, UP007
    ignore_roles: t.Optional[t.List[str]] = None  # noqa: UP006, UP007
    ignore_substitutions: t.Optional[t.List[str]] = None  # noqa: UP006, UP007
    ignore_languages: t.Optional[t.List[str]] = None  # noqa: UP006, UP007
    ignore_messages: t.Union[t.List[str], str, None] = None  # noqa: UP006, UP007


def _load_config_from_toml_file(
    toml_file: pathlib.Path,
    *,
    log_missing_section_as_warning: bool = True,
    warn_unknown_settings: bool = False,
) -> RstcheckConfigFile | None:
    """Load, parse and validate rstcheck config from a TOML file.

    .. warning::

        Needs tomli installed for python versions before 3.11!
        Use toml extra.

    :param toml_file: TOML file to load config from
    :param log_missing_section_as_warning: If a missing [tool.rstcheck] section should be logged at
        WARNING (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level;
        defaults to :py:obj:`True`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :raises ValueError: If the file is not a TOML file
    :raises FileNotFoundError: If the file is not found
    :return: instance of :py:class:`RstcheckConfigFile` or :py:obj:`None` on missing config section
        or ``NONE`` is passed as the config path.
    """
    _extras.install_guard_tomli(tomllib_imported=tomllib_imported)
    logger.debug("Try loading config from TOML file: '%s'.", toml_file)

    if toml_file.name == "NONE":
        logger.info("Config path is set to 'NONE'. No config file is loaded.")
        return None

    resolved_file = toml_file.resolve()

    if not resolved_file.is_file():
        logging.error("Config file is not a file: '%s'.", toml_file)
        msg = f"{resolved_file}"
        raise FileNotFoundError(msg)

    if resolved_file.suffix.casefold() != ".toml":
        logging.error("Config file is not a TOML file: '%s'.", toml_file)
        msg = "File is not a TOML file"
        raise ValueError(msg)

    with pathlib.Path(resolved_file).open("rb") as toml_file_handle:
        toml_dict = tomllib.load(toml_file_handle)

    rstcheck_section: t.Optional[dict[str, t.Any]] = toml_dict.get("tool", {}).get(  # noqa: UP007
        "rstcheck"
    )

    if rstcheck_section is None:
        if log_missing_section_as_warning:
            logger.warning(
                "Config file has no [tool.rstcheck] section: '%s'.",
                toml_file,
            )
            return None
        logger.info(
            "Config file has no [tool.rstcheck] section: '%s'.",
            toml_file,
        )
        return None

    if warn_unknown_settings:
        known_settings = _RstcheckConfigTOMLFile().model_dump().keys()
        unknown = [s for s in rstcheck_section if s not in known_settings]
        if unknown:
            logger.warning(
                "Unknown setting(s) %s found in file: '%s'.",
                unknown,
                toml_file,
            )

    config_values_checked = _RstcheckConfigTOMLFile(**rstcheck_section)
    return RstcheckConfigFile(**config_values_checked.model_dump())


def load_config_file(
    file_path: pathlib.Path,
    *,
    log_missing_section_as_warning: bool = True,
    warn_unknown_settings: bool = False,
) -> RstcheckConfigFile | None:
    """Load, parse and validate rstcheck config from a file.

    .. caution::

        If a TOML file is passed this function need tomli installed for python versions before 3.11!
        Use toml extra or install manually.

    :param file_path: File to load config from
    :param log_missing_section_as_warning: If a missing config section should be logged at
        WARNING (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level;
        defaults to :py:obj:`True`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :raises FileNotFoundError: If the file is not found
    :return: instance of :py:class:`RstcheckConfigFile` or :py:obj:`None` on missing config section
        or ``NONE`` is passed as the config path.
    """
    logger.debug("Try loading config file.")

    if file_path.name == "NONE":
        logger.info("Config path is set to 'NONE'. No config file is loaded.")
        return None

    if file_path.suffix.casefold() == ".toml":
        return _load_config_from_toml_file(
            file_path,
            log_missing_section_as_warning=log_missing_section_as_warning,
            warn_unknown_settings=warn_unknown_settings,
        )
    return _load_config_from_ini_file(
        file_path,
        log_missing_section_as_warning=log_missing_section_as_warning,
        warn_unknown_settings=warn_unknown_settings,
    )


def load_config_file_from_dir(
    dir_path: pathlib.Path,
    *,
    log_missing_section_as_warning: bool = False,
    warn_unknown_settings: bool = False,
) -> RstcheckConfigFile | None:
    """Search, load, parse and validate rstcheck config from a directory.

    Searches files from :py:data:`CONFIG_FILES` in the directory. If a file is found, try to load
    the config from it. If is has no config, search further.

    :param dir_path: Directory to search
    :param log_missing_section_as_warning: If a missing config section in a config file should be
        logged at WARNING (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level;
        defaults to :py:obj:`False`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :return: instance of :py:class:`RstcheckConfigFile` or
        :py:obj:`None` if no file is found or no file has a rstcheck section
        or ``NONE`` is passed as the config path.
    """
    logger.debug("Try loading config file from directory: '%s'.", dir_path)

    if dir_path.name == "NONE":
        logger.info("Config path is set to 'NONE'. No config file is loaded.")
        return None

    config = None

    for file_name in CONFIG_FILES:
        file_path = (dir_path / file_name).resolve()
        if file_path.is_file():
            config = load_config_file(
                file_path,
                log_missing_section_as_warning=(
                    log_missing_section_as_warning or (file_name == ".rstcheck.cfg")
                ),
                warn_unknown_settings=warn_unknown_settings,
            )
            if config is not None:
                break

    if config is None:
        logger.info(
            "No config section in supported config files found in directory: '%s'.",
            dir_path,
        )

    return config


def load_config_file_from_dir_tree(
    dir_path: pathlib.Path,
    *,
    log_missing_section_as_warning: bool = False,
    warn_unknown_settings: bool = False,
) -> RstcheckConfigFile | None:
    """Search, load, parse and validate rstcheck config from a directory tree.

    Searches files from :py:data:`CONFIG_FILES` in the directory. If a file is found, try to load
    the config from it. If is has no config, search further. If no config is found in the directory
    search its parents one by one.

    :param dir_path: Directory to search
    :param log_missing_section_as_warning: If a missing config section in a config file should be
        logged at ``WARNING`` (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level;
        defaults to :py:obj:`False`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :return: instance of :py:class:`RstcheckConfigFile` or
        :py:obj:`None` if no file is found or no file has a rstcheck section
        or ``NONE`` is passed as the config path.
    """
    logger.debug("Try loading config file from directory tree: '%s'.", dir_path)

    if dir_path.name == "NONE":
        logger.info("Config path is set to 'NONE'. No config file is loaded.")
        return None

    config = None

    search_dir = dir_path.resolve()

    while True:
        config = load_config_file_from_dir(
            search_dir,
            log_missing_section_as_warning=log_missing_section_as_warning,
            warn_unknown_settings=warn_unknown_settings,
        )

        if config is not None:
            break

        parent_dir = search_dir.parent.resolve()
        if parent_dir == search_dir:
            break
        search_dir = parent_dir

    if config is None:
        logger.info(
            "No config section in supported config files found in directory tree: '%s'.",
            dir_path,
        )

    return config


def load_config_file_from_path(
    path: pathlib.Path,
    *,
    search_dir_tree: bool = False,
    log_missing_section_as_warning_for_file: bool = True,
    log_missing_section_as_warning_for_dir: bool = False,
    warn_unknown_settings: bool = False,
) -> RstcheckConfigFile | None:
    """Analyse the path and call the correct config file loader.

    :param path: Path to load config file from; can be a file or directory
    :param search_dir_tree: If the directory tree should be searched;
        only applies if ``path`` is a directory;
        defaults to :py:obj:`False`
    :param log_missing_section_as_warning_for_file: If a missing config section in a config file
        should be logged at WARNING (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level when the
        given path is a file;
        defaults to :py:obj:`True`
    :param log_missing_section_as_warning_for_dir: If a missing config section in a config file
        should be logged at ``WARNING`` (:py:obj:`True`) or ``INFO`` (:py:obj:`False`) level when
        the given file is a direcotry;
        defaults to :py:obj:`False`
    :param warn_unknown_settings: If a warning should be logged for unknown settings in config file;
        defaults to :py:obj:`False`
    :raises FileNotFoundError: When the passed path is not found.
    :return: instance of :py:class:`RstcheckConfigFile` or
        :py:obj:`None` if no file is found or no file has a rstcheck section
        or ``NONE`` is passed as the config path.
    """
    logger.debug("Try loading config file from path: '%s'.", path)

    if path.name == "NONE":
        logger.info("Config path is set to 'NONE'. No config file is loaded.")
        return None

    resolved_path = path.resolve()

    if resolved_path.is_file():
        return load_config_file(
            resolved_path,
            log_missing_section_as_warning=log_missing_section_as_warning_for_file,
            warn_unknown_settings=warn_unknown_settings,
        )

    if resolved_path.is_dir():
        if search_dir_tree:
            return load_config_file_from_dir_tree(
                resolved_path,
                log_missing_section_as_warning=log_missing_section_as_warning_for_dir,
                warn_unknown_settings=warn_unknown_settings,
            )
        return load_config_file_from_dir(
            resolved_path,
            log_missing_section_as_warning=log_missing_section_as_warning_for_dir,
            warn_unknown_settings=warn_unknown_settings,
        )

    raise FileNotFoundError(2, "Passed config path not found.", path)


def merge_configs(
    config_base: RstcheckConfig,
    config_add: RstcheckConfig | RstcheckConfigFile,
    *,
    config_add_is_dominant: bool = True,
) -> RstcheckConfig:
    """Merge two configs into a new one.

    :param config_base: The base config to merge into
    :param config_add: The config that is merged into the ``config_base``
    :param config_add_is_dominant: If the ``config_add`` overwrites values of ``config_base``;
        defaults to :py:obj:`True`
    :return: New merged config
    """
    logger.debug("Merging configs.")
    sub_config: RstcheckConfig | RstcheckConfigFile = config_base
    sub_config_dict = sub_config.model_dump()
    for setting in dict(sub_config_dict):
        if sub_config_dict[setting] is None:
            del sub_config_dict[setting]

    dom_config: RstcheckConfig | RstcheckConfigFile = config_add
    dom_config_dict = dom_config.model_dump()
    for setting in dict(dom_config_dict):
        if dom_config_dict[setting] is None:
            del dom_config_dict[setting]

    if config_add_is_dominant is False:
        sub_config_dict, dom_config_dict = dom_config_dict, sub_config_dict

    merged_config_dict = {**sub_config_dict, **dom_config_dict}

    return RstcheckConfig(**merged_config_dict)



"""Sphinx helper functions."""

from __future__ import annotations

import contextlib
import logging
import pathlib
import tempfile
import typing as t

from . import _docutils, _extras

if _extras.SPHINX_INSTALLED:
    import sphinx.application
    import sphinx.domains.c
    import sphinx.domains.cpp
    import sphinx.domains.javascript
    import sphinx.domains.python
    import sphinx.domains.std
    import sphinx.util.docutils


logger = logging.getLogger(__name__)


def create_dummy_sphinx_app() -> sphinx.application.Sphinx:
    """Create a dummy sphinx instance with temp dirs."""
    logger.debug("Create dummy sphinx application.")
    with tempfile.TemporaryDirectory() as temp_dir:
        outdir = pathlib.Path(temp_dir) / "_build"
        return sphinx.application.Sphinx(
            srcdir=temp_dir,
            confdir=None,
            outdir=str(outdir),
            doctreedir=str(outdir),
            buildername="dummy",
            # NOTE: https://github.com/sphinx-doc/sphinx/issues/10483
            status=None,
        )


@contextlib.contextmanager
def load_sphinx_if_available() -> t.Generator[sphinx.application.Sphinx | None, None, None]:
    """Contextmanager to register Sphinx directives and roles if sphinx is available."""
    if _extras.SPHINX_INSTALLED:
        create_dummy_sphinx_app()
        # NOTE: Hack to prevent sphinx warnings for overwriting registered nodes; see #113
        sphinx.application.builtin_extensions = [
            e for e in sphinx.application.builtin_extensions if e != "sphinx.addnodes"  # type: ignore[assignment]
        ]

    yield None


def get_sphinx_directives_and_roles() -> tuple[list[str], list[str]]:
    """Return Sphinx directives and roles loaded from sphinx.

    :return: Tuple of directives and roles
    """
    _extras.install_guard("sphinx")

    sphinx_directives = list(sphinx.domains.std.StandardDomain.directives)
    sphinx_roles = list(sphinx.domains.std.StandardDomain.roles)

    for domain in [
        sphinx.domains.c.CDomain,
        sphinx.domains.cpp.CPPDomain,
        sphinx.domains.javascript.JavaScriptDomain,
        sphinx.domains.python.PythonDomain,
    ]:
        domain_directives = list(domain.directives)
        domain_roles = list(domain.roles)

        sphinx_directives += domain_directives + [
            f"{domain.name}:{item}" for item in domain_directives
        ]

        sphinx_roles += domain_roles + [f"{domain.name}:{item}" for item in domain_roles]

    sphinx_directives += list(
        sphinx.util.docutils.directives._directives  # type: ignore[attr-defined]  # noqa: SLF001
    )
    sphinx_roles += list(
        sphinx.util.docutils.roles._roles  # type: ignore[attr-defined]  # noqa: SLF001
    )

    return (sphinx_directives, sphinx_roles)


_DIRECTIVE_WHITELIST = ["code", "code-block", "sourcecode", "include"]
_ROLE_WHITELIST: list[str] = []


def filter_whitelisted_directives_and_roles(
    directives: list[str], roles: list[str]
) -> tuple[list[str], list[str]]:
    """Filter whitelisted directives and roles out of input.

    :param directives: Directives to filter
    :param roles: Roles to filter
    :return: Tuple of fitlered directives and roles
    """
    directives = list(filter(lambda d: d not in _DIRECTIVE_WHITELIST, directives))
    roles = list(filter(lambda r: r not in _ROLE_WHITELIST, roles))

    return (directives, roles)


def load_sphinx_ignores() -> None:  # pragma: no cover
    """Register Sphinx directives and roles to ignore."""
    _extras.install_guard("sphinx")
    logger.debug("Load sphinx directives and roles.")

    (directives, roles) = get_sphinx_directives_and_roles()
    (directives, roles) = filter_whitelisted_directives_and_roles(directives, roles)

    _docutils.ignore_directives_and_roles(directives, roles)



"""Helper types."""

from __future__ import annotations

import pathlib
import typing as t

SourceFileOrString = t.Union[pathlib.Path, t.Literal["<string>", "<stdin>"]]  # noqa: UP007
"""Path to source file or if it is a string then '<string>' or '<stdin>'."""


class LintError(t.TypedDict):
    """Dict with information about an linting error."""

    source_origin: SourceFileOrString
    line_number: int
    message: str


YieldedLintError = t.Generator[LintError, None, None]
"""Yielded version of type :py:class:`LintError`."""


class IgnoreDict(t.TypedDict):
    """Dict with ignore information."""

    messages: t.Pattern[str] | None
    languages: list[str]
    directives: list[str]
    roles: list[str]
    substitutions: list[str]


def construct_ignore_dict(
    messages: t.Pattern[str] | None = None,
    languages: list[str] | None = None,
    directives: list[str] | None = None,
    roles: list[str] | None = None,
    substitutions: list[str] | None = None,
) -> IgnoreDict:
    """Create an :py:class:`IgnoreDict` with passed values or defaults.

    :param messages: Value for :py:attr:`IgnoreDict.messages`;
        :py:obj:`None` results in an empty list; defaults to :py:obj:`None`
    :param directives: Value for :py:attr:`IgnoreDict.directives`;
        :py:obj:`None` results in an empty list; defaults to :py:obj:`None`
    :param roles: Value for :py:attr:`IgnoreDict.roles`;
        :py:obj:`None` results in an empty list; defaults to :py:obj:`None`
    :param substitutions: Value for :py:attr:`IgnoreDict.substitutions`;
        :py:obj:`None` results in an empty list; defaults to :py:obj:`None`
    :return: :py:class:`IgnoreDict` with passed values or defaults
    """
    return IgnoreDict(
        messages=messages,
        languages=languages if languages is not None else [],
        directives=directives if directives is not None else [],
        roles=roles if roles is not None else [],
        substitutions=substitutions if substitutions is not None else [],
    )


CheckerRunFunction = t.Callable[..., YieldedLintError]
"""Function to run checks.

Returned by :py:meth:`rstcheck_core.checker.CodeBlockChecker.create_checker`.
"""


class InlineConfig(t.TypedDict):
    """Dict with a config key and config value comming from a inline config comment."""

    key: str
    value: str


class InlineFlowControl(t.TypedDict):
    """Dict with a flow control value and line number comming from a inline config comment."""

    value: str
    line_number: int



"""Docutils helper functions."""

from __future__ import annotations

import importlib
import logging
import typing as t

import docutils.nodes
import docutils.parsers.rst.directives
import docutils.parsers.rst.roles
import docutils.writers

from . import _extras

logger = logging.getLogger(__name__)


class IgnoredDirective(docutils.parsers.rst.Directive):  # pragma: no cover
    """Stub for unknown directives."""

    has_content = True

    def run(self) -> list:  # type: ignore[type-arg]
        """Do nothing."""
        return []


def ignore_role(  # noqa: PLR0913
    name: str,  # noqa: ARG001
    rawtext: str,  # noqa: ARG001
    text: str,  # noqa: ARG001
    lineno: int,  # noqa: ARG001
    inliner: docutils.parsers.rst.states.Inliner,  # noqa: ARG001
    options: t.Mapping[str, t.Any] | None = None,  # noqa: ARG001
    content: t.Sequence[str] | None = None,  # noqa: ARG001
) -> tuple[
    t.Sequence[docutils.nodes.reference], t.Sequence[docutils.nodes.reference]
]:  # pragma: no cover
    """Stub for unknown roles."""
    return ([], [])


def clean_docutils_directives_and_roles_cache() -> None:  # pragma: no cover
    """Clean docutils' directives and roles cache by reloading their modules.

    Reloads:
    - :py:mod:`docutils.parsers.rst.directives`
    - :py:mod:`docutils.parsers.rst.roles`
    """
    logger.info("Reload module docutils.parsers.rst.directives/roles")
    importlib.reload(docutils.parsers.rst.directives)
    importlib.reload(docutils.parsers.rst.roles)


def ignore_directives_and_roles(directives: list[str], roles: list[str]) -> None:
    """Ignore directives and roles in docutils.

    :param directives: Directives to ignore
    :param roles: Roles to ignore
    """
    for directive in directives:
        docutils.parsers.rst.directives.register_directive(directive, IgnoredDirective)

    for role in roles:
        docutils.parsers.rst.roles.register_local_role(role, ignore_role)


class CodeBlockDirective(docutils.parsers.rst.Directive):
    """Code block directive."""

    has_content = True
    optional_arguments = 1

    def run(self) -> list[docutils.nodes.literal_block]:
        """Run directive.

        :return: Literal block
        """
        try:
            language = self.arguments[0]
        except IndexError:
            language = ""
        code = "\n".join(self.content)
        literal = docutils.nodes.literal_block(code, code)
        literal["classes"].append("code-block")
        literal["language"] = language
        return [literal]


def register_code_directive(
    *,
    ignore_code_directive: bool = False,
    ignore_codeblock_directive: bool = False,
    ignore_sourcecode_directive: bool = False,
) -> None:
    """Optionally register code directives.

    :param ignore_code_directive: If "code" directive should be ignored,
        so that the code block will not be checked; defaults to :py:obj:`False`
    :param ignore_codeblock_directive: If "code-block" directive should be ignored,
        so that the code block will not be checked; defaults to :py:obj:`False`
    :param ignore_sourcecode_directive: If "sourcecode" directive should be ignored,
        so that the code block will not be checked; defaults to :py:obj:`False`
    """
    if not _extras.SPHINX_INSTALLED:
        if ignore_code_directive is False:
            logger.debug("Register custom directive for 'code'.")
            docutils.parsers.rst.directives.register_directive("code", CodeBlockDirective)
        # NOTE: docutils maps `code-block` and `sourcecode` to `code`
        if ignore_codeblock_directive is False:
            logger.debug("Register custom directive for 'code-block'.")
            docutils.parsers.rst.directives.register_directive("code-block", CodeBlockDirective)
        if ignore_sourcecode_directive is False:
            logger.debug("Register custom directive for 'sourcecode'.")
            docutils.parsers.rst.directives.register_directive("sourcecode", CodeBlockDirective)



"""Inline config comment functionality."""

from __future__ import annotations

import functools
import logging
import re
import typing as t

from . import types

logger = logging.getLogger(__name__)


RSTCHECK_CONFIG_COMMENT_REGEX = re.compile(r"\.\. rstcheck: (.*)=(.*)$")
VALID_INLINE_CONFIG_KEYS = (
    "ignore-directives",
    "ignore-roles",
    "ignore-substitutions",
    "ignore-languages",
)
ValidInlineConfigKeys = t.Literal[
    "ignore-directives", "ignore-roles", "ignore-substitutions", "ignore-languages"
]

RSTCHECK_FLOW_CONTROL_COMMENT_REGEX = re.compile(r"\.\. rstcheck: ([a-z-]*)$")
VALID_INLINE_FLOW_CONTROLS = ("ignore-next-code-block",)


@functools.lru_cache
def get_inline_config_from_source(
    source: str, source_origin: types.SourceFileOrString, *, warn_unknown_settings: bool = False
) -> list[types.InlineConfig]:
    """Get rstcheck inline configs from source.

    Unknown configs are ignored.

    :param source: Source to get config from
    :param source_origin: Origin of the source with the inline ignore comments
    :param warn_unknown_settings: If a warning should be logged on unknown settings;
        defaults to :py:obj:`False`
    :return: A list of inline configs
    """
    configs: list[types.InlineConfig] = []
    for idx, line in enumerate(source.splitlines()):
        match = RSTCHECK_CONFIG_COMMENT_REGEX.search(line)
        if match is None:
            continue

        key = match.group(1).strip()
        value = match.group(2).strip()

        if key not in VALID_INLINE_CONFIG_KEYS:
            if warn_unknown_settings:
                logger.warning(
                    "Unknown inline config '%s' found. Source: '%s' at line %s",
                    key,
                    source_origin,
                    idx + 1,
                )
            continue

        configs.append(types.InlineConfig(key=key, value=value))

    return configs


def _filter_config_and_split_values(
    target_config: ValidInlineConfigKeys,
    source: str,
    source_origin: types.SourceFileOrString,
    *,
    warn_unknown_settings: bool = False,
) -> t.Generator[str, None, None]:
    """Get specified configs and comma split them.

    :param target_config: Config target to filter for
    :param source: Source to get config from
    :param source_origin: Origin of the source with the inline ignore comments
    :param warn_unknown_settings: If a warning should be logged on unknown settings;
        defaults to :py:obj:`False`
    :return: None
    :yield: Single values for the ``target_config``
    """
    inline_configs = get_inline_config_from_source(
        source, source_origin, warn_unknown_settings=warn_unknown_settings
    )
    for inline_config in inline_configs:
        if inline_config["key"] == target_config:
            for language in inline_config["value"].split(","):
                yield language.strip()


def find_ignored_directives(
    source: str, source_origin: types.SourceFileOrString, *, warn_unknown_settings: bool = False
) -> t.Generator[str, None, None]:
    """Search the rst source for rstcheck inline ignore-directives comments.

    Directives are ignored via comment.

    For example, to ignore directive1, directive2, and directive3:

    .. testsetup::

        from rstcheck_core.inline_config import find_ignored_directives

    >>> list(find_ignored_directives('''
    ... Example
    ... =======
    ...
    ... .. rstcheck: ignore-directives=directive1,directive3
    ...
    ... .. rstcheck: ignore-directives=directive2
    ... ''', "<string>"))
    ['directive1', 'directive3', 'directive2']

    :param source: Rst source code
    :param source_origin: Origin of the source with the inline ignore comments
    :return: None
    :yield: Found directives to ignore
    """
    yield from _filter_config_and_split_values(
        "ignore-directives", source, source_origin, warn_unknown_settings=warn_unknown_settings
    )


def find_ignored_roles(
    source: str, source_origin: types.SourceFileOrString, *, warn_unknown_settings: bool = False
) -> t.Generator[str, None, None]:
    """Search the rst source for rstcheck inline ignore-roles comments.

    Roles are ignored via comment.

    For example, to ignore role1, role2, and role3:

    .. testsetup::

        from rstcheck_core.inline_config import find_ignored_roles

    >>> list(find_ignored_roles('''
    ... Example
    ... =======
    ...
    ... .. rstcheck: ignore-roles=role1,role3
    ...
    ... .. rstcheck: ignore-roles=role2
    ... ''', "<string>"))
    ['role1', 'role3', 'role2']

    :param source: Rst source code
    :param source_origin: Origin of the source with the inline ignore comments
    :return: None
    :yield: Found roles to ignore
    """
    yield from _filter_config_and_split_values(
        "ignore-roles", source, source_origin, warn_unknown_settings=warn_unknown_settings
    )


def find_ignored_substitutions(
    source: str, source_origin: types.SourceFileOrString, *, warn_unknown_settings: bool = False
) -> t.Generator[str, None, None]:
    """Search the rst source for rstcheck inline ignore-substitutions comments.

    Substitutions are ignored via comment.

    For example, to ignore substitution1, substitution2, and substitution3:

    .. testsetup::

        from rstcheck_core.inline_config import find_ignored_substitutions

    >>> list(find_ignored_substitutions('''
    ... Example
    ... =======
    ...
    ... .. rstcheck: ignore-substitutions=substitution1,substitution3
    ...
    ... .. rstcheck: ignore-substitutions=substitution2
    ... ''', "<string>"))
    ['substitution1', 'substitution3', 'substitution2']

    :param source: Rst source code
    :param source_origin: Origin of the source with the inline ignore comments
    :return: None
    :yield: Found substitutions to ignore
    """
    yield from _filter_config_and_split_values(
        "ignore-substitutions", source, source_origin, warn_unknown_settings=warn_unknown_settings
    )


def find_ignored_languages(
    source: str, source_origin: types.SourceFileOrString, *, warn_unknown_settings: bool = False
) -> t.Generator[str, None, None]:
    """Search the rst source for rstcheck inline ignore-languages comments.

    Languages are ignored via comment.

    For example, to ignore C++, JSON, and Python:

    .. testsetup::

        from rstcheck_core.inline_config import find_ignored_languages

    >>> list(find_ignored_languages('''
    ... Example
    ... =======
    ...
    ... .. rstcheck: ignore-languages=cpp,json
    ...
    ... .. rstcheck: ignore-languages=python
    ... ''', "<string>"))
    ['cpp', 'json', 'python']

    :param source: Rst source code
    :param source_origin: Origin of the source with the inline ignore comments
    :return: None
    :yield: Found languages to ignore
    """
    yield from _filter_config_and_split_values(
        "ignore-languages", source, source_origin, warn_unknown_settings=warn_unknown_settings
    )


@functools.lru_cache
def get_inline_flow_control_from_source(
    source: str, source_origin: types.SourceFileOrString, *, warn_unknown_settings: bool = False
) -> list[types.InlineFlowControl]:
    """Get rstcheck inline flow control from source.

    Unknown flow controls are ignored.

    :param source: Source to get config from
    :param source_origin: Origin of the source with the inline flow control
    :param warn_unknown_settings: If a warning should be logged on unknown settings;
        defaults to :py:obj:`False`
    :return: A list of inline flow controls
    """
    configs: list[types.InlineFlowControl] = []
    for idx, line in enumerate(source.splitlines()):
        match = RSTCHECK_FLOW_CONTROL_COMMENT_REGEX.search(line)
        if match is None:
            continue

        value = match.group(1).strip()
        line_number = idx + 1

        if value not in VALID_INLINE_FLOW_CONTROLS:
            if warn_unknown_settings:
                logger.warning(
                    "Unknown inline flow control '%s' found. Source: '%s' at line %s",
                    value,
                    source_origin,
                    line_number,
                )
            continue

        configs.append(types.InlineFlowControl(value=value, line_number=line_number))

    return configs


def find_code_block_ignore_lines(
    source: str,
    source_origin: types.SourceFileOrString,
    *,
    warn_unknown_settings: bool = False,
) -> t.Generator[int, None, None]:
    """Get lines of ``ignore-next-code-block`` flow control comments.

    :param source: Source to get config from
    :param source_origin: Origin of the source with the inline ignore comments
    :param warn_unknown_settings: If a warning should be logged on unknown settings;
        defaults to :py:obj:`False`
    :return: None
    :yield: Single values for the ``target_config``
    """
    flow_controls = get_inline_flow_control_from_source(
        source, source_origin, warn_unknown_settings=warn_unknown_settings
    )
    for flow_control in flow_controls:
        if flow_control["value"] == "ignore-next-code-block":
            yield flow_control["line_number"]
